/*
    Copyright (C) 2003, Patrick Ahlbrecht

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <getopt.h>
#include <regex.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include "util.h"

// Inputbuffersize for fread() - 150kb should be sane for most pages
#define BUFSIZE 1024*150

// Set program version               
char* version=VERSION;

// Broken up URL
struct link {
  char* asis;		// Original URL (not broken down)
  char* prot;		// Protocol part
  char* host;		// Address part
  char* port;		// Port part (This is char* to spare conversion)
  char* path;		// Path part
  char* args;		// Optional Arguments
  char* mail;		// Only set, if the URL is an emailaddress
  char* user;		// UID credentials
  char* pass;		// PASS credentials
};

// Storage for URL's found
struct link** links=NULL;

// Number of URL's in storage
int linkc=0;


/**
 * Store URL's found
 * @param u URL to store
 */
void store(const struct link* u) {
  if (u==NULL) return;
  linkc++;
  links=(struct link**)realloc(links,linkc*sizeof(struct link*)+1);
  oomtest(links);
  links[linkc-1]=(struct link*)u;
  links[linkc]=NULL;
}


/**
 * Process raw text
 * @param txt the complete text
 * @param begin index of the first URL character
 * @param end index of the last URL character
 * @return an link struct derived from the input
 */
struct link* txt2link(const char* txt,const int begin,const int end) {
  int idx=0;
  int mark=0;
  
  struct link* u=(struct link*)calloc(1,sizeof(struct link));
  oomtest(u);
  
  u->asis=cpyTxt(txt,begin,end);
  
  if(u->asis==NULL) return u;
  
  // Let's try to determine the protocol first
  while(u->asis[idx]!='\0' && u->asis[idx]!='\0' && u->asis[idx]!=':') idx++;
  if(u->asis[idx+1]=='/' && u->asis[idx+2]=='/') {
    u->prot=cpyTxt(u->asis,0,idx);
    idx=idx+3;
  }
  
  // Of course, "mailto:" is something completly different...
  if(!strncasecmp("mailto:",u->asis,idx)) {
    u->prot=cpyTxt(u->asis,0,idx);
    u->mail=cpyTxt(u->asis,idx+1,strlen(u->asis));
    return u;
  }
  
  // A hostaddress requires a protocol being specified...
  if (u->prot!=NULL) {
    char** tmp;
    char** tmp2;
    char* s;
    
    mark=idx;
    while(u->asis[idx]!='\0' && u->asis[idx]!='/') idx++;
    if (mark==idx) return u;
    s=cpyTxt(u->asis,mark,idx);
    tmp=splitString(s,3,'@');
    free(s);
    
    if (tmp[1]!=NULL) {
      tmp2=splitString(tmp[0],2,':');
      u->user=cpyStr(tmp2[0]);
      u->pass=cpyStr(tmp2[1]);
      freeList((void**)tmp2);
      tmp2=splitString(tmp[1],2,':');
      u->host=cpyStr(tmp2[0]);
      u->port=cpyStr(tmp2[1]);
      freeList((void**)tmp2);
    }
    else {
      tmp2=splitString(tmp[0],2,':');
      u->host=cpyStr(tmp2[0]);
      u->port=cpyStr(tmp2[1]);
      freeList((void**)tmp2);
    }
    
    freeList((void**)tmp);
    
  }
  
  // Now let's figure out the path argument
  if(u->prot==NULL) idx=0;
  mark=idx;
  while(u->asis[idx]!='\0' && u->asis[idx]!='?') idx++;
  u->path=cpyTxt(u->asis,mark,idx);
  
  // Finally the arguments
  u->args=cpyTxt(u->asis,idx,strlen(u->asis));

  return u;
}

/**
 * Parse an HTML document, get all URLs
 * @param txt pointer to the input text
 */
void parse(const char* txt) {
  size_t idx=0;
  int open=FALSE;
  size_t max;
  
  if (txt==NULL) return;
  max=strlen(txt);	// calling strlen() each loop kills performance
  
  while(idx<=max) {
    if (txt[idx]=='<' && !open) open=TRUE;
    if (txt[idx]=='>' && open) open=FALSE;
    
    if ((!strncasecmp(txt+idx,"HREF",4)) && open) {
      int off;

      idx=idx+4;
      while(txt[idx]==' ' || txt[idx]=='=' || txt[idx]=='\"') idx++;
      off=idx;
      while (txt[off]!='\0' && txt[off]!='>' && txt[off]!='\"') off++;
      if (txt[off]=='\"' || txt[off]=='>') store(txt2link(txt,idx,off));
    }
    idx++;
  }
}


/**
 * Print usage information and exit
 * @param name argv[0] of main.
 */
void printUsage(const char* name) {
  char* tmp=(char*)basename((char*)name);
    
  printf("%s - Extract links from HTML pages\n\
  \nOptions:\n\n\
  -h\t\tPrint usage\n\
  -i\t\tInverse match (use with \"-r\")\n\
  -g <file>\tGenerate output, using <file> as template\n\
  -f <file>\tRead input from <file>\n\
  -r <regex>\tOnly print links matching <regex>\n\
  -a <macro>\tApply regex to <macro> (default: \"=asis\")\n\
  -t <string>\tUse <string> as output template (see \"Macros:\" below)\n\
  \nMacros:\n\n\
  =asis\t\tComplete link (not broken into parts)\n\
  =prot\t\tProtocol (in case link is an URL)\n\
  =host\t\tHostaddress (in case link is an URL)\n\
  =port\t\tPort number (in case link is an URL)\n\
  =path\t\tPath the link points to\n\
  =args\t\tOptional arguments\n\
  =mail\t\tEmailadress (in case link is an emailaddress link)\n\
  =user\t\tloginname (in case link contains credentials)\n\
  =pass\t\tpassword (in case link contains credentials)\n\
  \\n\t\tA newline character\n\
  \\t\t\tA Tabulator character\n\n\
  Not every macro may be defined for every link. If a macro is undefined,\n\
  it will be surpressed in the output.\n\
  \n",tmp);
  exit(EXIT_SUCCESS);
}

/**
 * Obvious ;-)
 */
int main(int argc,char**argv) {
  int ch;
  char* format;
  char* buffer=NULL;
  FILE* in=stdin;
  char* reg=NULL;
  char* to="=asis";
  int im=FALSE;
  
  format=getenv("FSTRING");
  if (format==NULL) format="=asis\n";
  
  links=(struct link**)calloc(1,sizeof(struct link*));
  oomtest(links);
  
  while ((ch = getopt(argc, argv, "ht:g:f:r:a:i")) !=-1) {
    switch(ch) {
      case 'h': { printUsage(argv[0]); break; }
      case 't': { format=optarg; break; }
      case 'g': { 
        format=readFile(optarg);
        if(format==NULL) {
          perror(optarg);
          return EXIT_FAILURE;
        }
        break;
      }
      case 'r': { reg=optarg; break; }
      case 'a': { to=optarg; break; }
      case 'i': { im=TRUE; break;}
      case 'f': {
        in=fopen(optarg,"r");
        if(in==NULL) {
          perror(optarg);
          exit(EXIT_FAILURE);
        }
        break;
      }
      
      default: printUsage(argv[0]);
    }
  }
  
  ch=1;
  
  while(!feof(in)) {
    char* buf;
    
    buf=(char*)calloc(BUFSIZE,sizeof(char));
    oomtest(buf);
    fread(buf,sizeof(char),BUFSIZE-1,in);
    buffer=(char*)realloc(buffer,sizeof(char)*BUFSIZE*ch);
    oomtest(buffer);
    strcat(buffer,buf);
    ch++;
    free(buf);
  }
  
  parse(buffer);
  free(buffer);
  
  ch=0;
  while (links[ch]!=NULL) {
    char* tmp=cpyStr(format);
    tmp=expand(tmp,"\\n","\n");
    tmp=expand(tmp,"\\t","\t");
    tmp=expand(tmp,"=asis",links[ch]->asis);
    tmp=expand(tmp,"=prot",links[ch]->prot);
    tmp=expand(tmp,"=host",links[ch]->host);
    tmp=expand(tmp,"=port",links[ch]->port);
    tmp=expand(tmp,"=path",links[ch]->path);
    tmp=expand(tmp,"=args",links[ch]->args);
    tmp=expand(tmp,"=mail",links[ch]->mail);
    tmp=expand(tmp,"=user",links[ch]->user);
    tmp=expand(tmp,"=pass",links[ch]->pass);
    if(reg==NULL) printf(tmp);
    
    else {
      regex_t regex;
      int reti;
      char* target;
      
      if (regcomp(&regex, reg, 0)) {
        printf("Invalid regex: %s\n",reg);
        exit(EXIT_FAILURE);
      }
      
      if(!strcasecmp(to,"=asis"))
        target=links[ch]->asis;
      else if (!strcasecmp(to,"=prot"))
        target=links[ch]->prot;
      else if (!strcasecmp(to,"=host"))
        target=links[ch]->host;
      else if (!strcasecmp(to,"=port"))
        target=links[ch]->port;
      else if (!strcasecmp(to,"=path"))
        target=links[ch]->path;
      else if (!strcasecmp(to,"=args"))
        target=links[ch]->path;
      else if (!strcasecmp(to,"=mail"))
        target=links[ch]->mail;
      else if (!strcasecmp(to,"=user"))
        target=links[ch]->user;
      else if (!strcasecmp(to,"=pass"))
        target=links[ch]->pass;
      else {
        printf("Unknown macro for regex: %s\n",to);
        exit(EXIT_FAILURE);
      }
      
      reti=REG_NOMATCH;
      
      if(target!=NULL) {
        reti=regexec(&regex, target, 0, NULL, 0);
      }
      
      if (reti != REG_NOMATCH && reti!=0) {
        printf("Regex failed: %s",reg);
        exit(EXIT_FAILURE);
      }
      
      if (reti==0 && im==FALSE) printf(tmp);
      if (reti==REG_NOMATCH && im==TRUE) printf(tmp);
      
      regfree(&regex);
    }
    
    ch++;
  }
  
  return EXIT_SUCCESS;
}
