#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <gdbm.h>
#include <time.h>
#include <errno.h>
#include <sys/stat.h>
#include <regex.h>

#include "util.h"

// Actions to take
#define STORE 1
#define DELET 2
#define QUERY 4
#define APPEN 8
#define LISTE 16
#define ERASE 32

// For debugging purposes
char* version=VERSION;


/**
 * Open GDBM database file
 * @param fname name of  the file to open
 * @param mode File open mode (directly passed through to gdbm_open)
 * @return either a filepointer to the opened database file or NULL
 * (errno is left unchanged).
 */
GDBM_FILE dbOpen( char* fname, int mode) {
  GDBM_FILE dbf;
  int retrycount=0;
  struct stat fs;
  
  if (fname==NULL) return NULL;
  
  if ( (mode==GDBM_READER) && (stat(fname,&fs)==-1) ) return NULL;
  
  do {
    dbf=gdbm_open(fname,0,mode,0666&~umask(0),NULL);
    if (errno==EACCES) return NULL;
    if (dbf!=NULL) break;
    retrycount++;
    sleep(1+(int) (10.0*rand()/(RAND_MAX+1.0)));
  }
  while ((errno == EAGAIN) && (retrycount<=10));
  
  return dbf;
}

/**
 * Close a GDBM file
 * @param dbf_ptr Pointer to the database file (will be set to NULL 
 * afterwards).
 */
void dbClose(GDBM_FILE dbf_ptr) {
  if (dbf_ptr==NULL) return;
  gdbm_sync(dbf_ptr);
  gdbm_close(dbf_ptr);
  dbf_ptr=NULL;
}

/**
 * Query whether or not a certain key exists in the database
 * @param str_ptr the string to test
 * @param dbf_ptr Pointer to the database file
 * @return TRUE key exists.
 */
int dbContains(const char* str_ptr, GDBM_FILE dbf_ptr) {
  datum key;
  
  if ((str_ptr==NULL) || (dbf_ptr==NULL)) return FALSE;
  key.dptr=(char*)str_ptr;
  key.dsize=(int)strlen(key.dptr)+sizeof(char);
  if (gdbm_exists(dbf_ptr,key)) return TRUE;
  return FALSE;
}

/**
 * Format entry
 * @param key the key value
 * @param val the value of the key
 * @param fmt format string
 * @return the formated entry
 */
char* formatEntry(const datum key, const datum val, const char* fmt) {
  char* tmp=cpyStr(fmt);
  oomtest(tmp);
  tmp=expand(tmp,"=key",key.dptr);
  tmp=expand(tmp,"=val",val.dptr);
  tmp=expand(tmp,"\\n","\n");
  tmp=expand(tmp,"\\t","\t");
  return tmp;
}


/**
 * Print usage and exit
 * @param name argv[0] of main
 */
void printUsage(const char* name) {
  printf("%s [OPTION] ...\n\
  \nRemember text strings in a hash file\n\n\
  -h\t\tPrint usage\n\
  -i\t\tInverse regular expression matching\n\
  -s <key>\tStore <key> (overwrite if already stored)\n\
  -a <key>\tAdd <key> (exit with failure, if already stored)\n\
  -d <key>\tDelete <key>\n\
  -v <val>\tOptionally associate <val> with a <key>\n\
  -l <file>\tList all strings stored in <file>\n\
  -f <file>\tUse <file> to store strings (default: %s.db)\n\
  -e <regex>\tErase entries, where the format string matches <regex>\n\
  -q <regex>\tQuery entries where the format string matches <regex>\n\
  -t <string>\tUse <string> as output template (default: \"=key\\n\")\n\
  \nThe following macros will be expanded in the format string:\n\n\
  =key\tKey name\n\
  =val\tValue associated with a key (if available)\n\
  \\n\tA newline character\n\
  \\t\tA tabulator character\n\
  \n",name,name);
  exit(EXIT_SUCCESS);
}

int main(int argc, char** argv) {
  int ch;
  char* bname;
  datum key;
  datum val;
  char* file;
  char* fstring;
  char* regex;
  int action=0;
  int inv=FALSE;
  GDBM_FILE dbf=NULL; 
  
  fstring=getenv("FSTRING");
  if(fstring==NULL) fstring="=key\n";
  
  bname=(char*)basename(argv[0]);
  file=expand(cpyStr("name.db"),"name",bname);
  
  val.dptr="";
  val.dsize=1;
  
  while ((ch = getopt(argc, argv, "his:d:q:v:f:l:t:a:e:")) !=-1) {
    switch(ch) {
      case 'h': { printUsage(bname); break; }
      case 'i': { inv=TRUE; break; }
      case 's': { 
        key.dptr=optarg;
        key.dsize=strlen(optarg)+1;
        action=action|STORE; 
        break; 
      }
      case 'd': {
        key.dptr=optarg;
        key.dsize=strlen(optarg)+1;
        action=action|DELET;
        break;
      }
      case 'q': {
        regex=optarg;
        action=action|QUERY;
        break;
      }
      case 'v': {
        val.dptr=optarg;
        val.dsize=strlen(optarg)+1;
        break;
      }
      case 'f': { file=optarg; break; }
      case 'l': { 
        action=action|LISTE;
        file=optarg;
        break;
      }
      case 't': { fstring=optarg; break; }
      case 'a': {
        key.dptr=optarg;
        key.dsize=strlen(optarg)+1;
        action=action|APPEN;
        break;
      }
      case 'e': {
        action=action|ERASE;
        regex=optarg;
        break;
      }
    }
  }
  
  if(action==0) printUsage(bname);
  
  if ((action&STORE)==STORE && (action&APPEN)==APPEN  || (action&DELET)==DELET && (action&ERASE)==ERASE) {
    fprintf(stderr,"Conflicting arguments\n");
    return EXIT_FAILURE;
  }
  
  if ((action&STORE)==STORE) {
    dbf=dbOpen(file,GDBM_WRCREAT);
    if(dbf==NULL) {
      perror("Could not open database file");
      return EXIT_FAILURE;
    }
    gdbm_store(dbf,key,val,GDBM_REPLACE);
    dbClose(dbf);
  }
  
  if ((action&APPEN)==APPEN) {
    dbf=dbOpen(file,GDBM_WRCREAT);
    if(dbf==NULL) {
      perror("Could not open database file");
      return EXIT_FAILURE;
    }
    if (gdbm_store(dbf,key,val,GDBM_INSERT)) {
      //fprintf(stderr,"Already stored\n");
      dbClose(dbf);
      return EXIT_FAILURE;
    }
    dbClose(dbf);
  }
  
  if ((action&DELET)==DELET) {
    dbf=dbOpen(file,GDBM_WRITER);
    if(dbf==NULL) {
      perror("Could not open database file");
      return EXIT_FAILURE;
    }
    if (gdbm_delete(dbf,key)==-1) {
      dbClose(dbf);
      return EXIT_FAILURE;
    }
    dbClose(dbf);
  }
  
  if ((action&QUERY)==QUERY) {
    datum v;
    datum k;
    datum n;
    char* tmp;
    regex_t reg;
    int reti;
    int ret=EXIT_FAILURE;
    
    dbf=dbOpen(file,GDBM_READER);
    if(dbf==NULL) {
      perror("Could not open database file");
      return EXIT_FAILURE;
    }
    
    k=gdbm_firstkey(dbf);
    
    while(k.dptr!=NULL) {
      if (regcomp(&reg, regex, 0)) {
        fprintf(stderr,"Invalid regex: %s\n",regex);
        dbClose(dbf);
        return EXIT_FAILURE;
      }
      n=gdbm_nextkey(dbf,k);
      
      v=gdbm_fetch(dbf,k);
      tmp=formatEntry(k,v,fstring);
      reti=regexec(&reg, tmp, 0, NULL, 0);
      
      if ((reti==0 && inv==FALSE) || (reti==REG_NOMATCH && inv==TRUE)) {
        printf("%s",tmp);
        ret=EXIT_SUCCESS;
      }
      
      free(tmp);
      free(v.dptr);
      free(k.dptr);
      regfree(&reg);
      k=n;
    }
    dbClose(dbf);
    if(ret==EXIT_FAILURE) return ret;
  }
  
  if((action&LISTE)==LISTE) {
    datum k;
    datum n;
    datum v;
    char* tmp;
    
    dbf=dbOpen(file,GDBM_READER);
    
    if(dbf==NULL) {
      perror("Could not open database file");
      return EXIT_FAILURE;
    }
 
    k=gdbm_firstkey(dbf);
    
    while(k.dptr!=NULL) {
      n=gdbm_nextkey(dbf,k);
      v=gdbm_fetch(dbf,k);
      tmp=formatEntry(k,v,fstring);
      printf("%s",tmp);
      free(tmp);
      free(v.dptr);
      free(k.dptr);
      k=n;
    }
    dbClose(dbf);
  }
  
  
  if((action&ERASE)==ERASE) {
    datum k;
    datum v;
    datum n;
    datum** d;
    char* tmp;
    regex_t reg;
    int reti;
    int c=0;
    
    dbf=dbOpen(file,GDBM_WRITER);
    
    if(dbf==NULL) {
      perror("Could not open database file");
      return EXIT_FAILURE;
    }
    d=(datum**)calloc(1,sizeof(datum));
    oomtest(d);
    k=gdbm_firstkey(dbf);
    
    while(k.dptr!=NULL) {
      if (regcomp(&reg, regex, 0)) {
        fprintf(stderr,"Invalid regex: %s\n",regex);
        dbClose(dbf);
        return EXIT_FAILURE;
      }
      n=gdbm_nextkey(dbf,k);
      
      v=gdbm_fetch(dbf,k);
      tmp=formatEntry(k,v,fstring);
      reti=regexec(&reg, tmp, 0, NULL, 0);
      
      if ((reti==0 && inv==FALSE) || (reti==REG_NOMATCH && inv==TRUE)) {
        datum* t;
        
        c++;
        d=(datum**)realloc(d,(c+1)*sizeof(datum*));
        t=(datum*)malloc(sizeof(datum));
        oomtest(d);
        oomtest(t);
        memcpy(t,&k,sizeof(datum));
        d[c-1]=t;
        d[c]=NULL;
      } else free(k.dptr);
      
      free(tmp);
      free(v.dptr);
      regfree(&reg);
      k=n;
    }
    
    c=0;
    while(d[c]!=NULL){
      gdbm_delete(dbf,*(d[c]));
      free(k.dptr);
      free(d[c]);
      c++;
    }
    free(d);
    gdbm_reorganize(dbf);
    dbClose(dbf);
  }
  
  return EXIT_SUCCESS;
}
